/*
 * etsh - an enhanced port of the Version 6 (V6) Thompson shell
 */
/*-
 * Copyright (c) 2004-2019
 *	Jeffrey Allen Neitzel <jan (at) etsh (dot) nl>.
 *	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JEFFREY ALLEN NEITZEL ``AS IS'', AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JEFFREY ALLEN NEITZEL BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	@(#)$Id: lib.h,v 1.6 2019/02/14 16:23:15 jneitzel Exp $
 */

#ifndef	LIB_H
#define	LIB_H

#define	DO_TRIM(k)	((k) != SBI_CD && (k) != SBI_CHDIR)

/* etsh.c */
extern	const char	*name;		/* $0 - shell command name          */
extern	bool		is_noexec;	/* not executable file flag         */
extern	bool		prompt_on;	/* `prompt' on/off flag; true == on */
extern	bool		prompt_dbg;	/* `prompt' debug  flag; true == on */
extern	bool		no_lnum;	/* no line number flag              */

/* lib.c */
void		error(int, /*@null@*/ const char *);
void		error1(int, /*@null@*/ const char *,
		       /*@null@*/ const char *);
void		error2(int, /*@null@*/ const char *,
		       /*@null@*/ const char *, /*@null@*/ const char *);
void		xfree(/*@null@*/ /*@only@*/ void *);
/*@maynotreturn@*/ /*@out@*/
void		*xmalloc(size_t);
void		vfree(/*@null@*/ char **);
/*@maynotreturn@*/
void		*xrealloc(/*@only@*/ void *, size_t);
/*@maynotreturn@*/
char		*xstrdup(/*@null@*/ const char *);
void		prompt_write(void);

#ifndef	S_SPLINT_S
char		*atrim(/*@notnull@*/ /*@returned@*/ u_char *, bool);
#endif

/*@null@*/
char		*gchar(/*@returned@*/ const char *);
/*@maynotreturn@*/ /*@null@*/ /*@only@*/
const char	**glob(enum sbikey, /*@only@*/ char **);

#endif	/* !LIB_H */
