#
# @(#)$Id: config.rc,v 1.2 2018/07/02 02:34:07 jneitzel Exp $
#
# Please specify your preferred default binary names for
# etsh (etsh_binary_name) and tsh (tsh_binary_name) below.
#
# Notice that etsh no longer conflicts w/ omake & its osh,
# as "etsh" != "osh".
#

#
# default
#
etsh_binary_name='etsh'
 tsh_binary_name='tsh'
